import sys
import struct
from bidict import bidict
from .utils import get_bytes_to_represent, get_binary_format_symbol, get_symbol_to_represent
from .exceptions import ImproperlyConfigured
from .message import MessageEnumMixin, MessageStringMixing, MessageBase


class MessageFactory(object):
    """
    Factory generates message classes based on a schema provided
    It also keeps track of message ids using message_types
    """

    msg_classes_by_name = {}
    msg_classes_by_id = {}
    binary_types = {
        'bool': '?',
        'byte': 'b',
        'ubyte': 'B',
        'char': 'c',
        'short': 'h',
        'ushort': 'H',
        'int': 'i',
        'uint': 'I',
        'int64': 'q',
        'uint64': 'Q',
        'float': 'f',
        'double': 'd'
    }

    def get_binary_format(self, msg_schema):
        """
        Build binary format for give message schema.
        Used internally to populate "binary_format" field on
        the message class.

        Output of thes method cannot be used directly with struct
        as it needs to be processed by the message instance first
        for handling dynamic-length strings.

        """
        fields = list(msg_schema['format'].keys())
        fields.sort()
        binary_format = '!'  # we always use network (big-endian) byte order
        binary_format += self.id_binary_format

        for field in fields:
            if(msg_schema['format'][field] == 'string'):
                binary_format += 'I{}s'
            elif(msg_schema['format'][field] == 'enum'):
                try:
                    binary_format += get_symbol_to_represent(
                        len(msg_schema['enums'][field])
                    )
                except Exception:
                    raise ImproperlyConfigured(
                        '''Enum field can contain the
                        maximum number {} possible values'''.format(
                            sys.maxsize
                        )
                    )
            else:
                try:
                    field_type = msg_schema['format'][field]
                    binary_format += self.binary_types[field_type]
                except KeyError:
                    raise ImproperlyConfigured(
                        "Unknown field type {}".format(
                            msg_schema['format'][field]
                        )
                    )

        return binary_format

    def get(self, id_or_name):
        """
        Convinience method to return message class for given name or id
        If number is given, we assume id, otheriwse - name.
        """
        if(type(id_or_name) == int):
            return self.get_by_id(id_or_name)
        else:
            return self.get_by_name(id_or_name)

    def get_by_name(self, name):
        """ Return message class given a name (string) """
        try:
            return self.msg_classes_by_name[name]
        except KeyError:
            raise KeyError(
                'No message by the name of {} found in the schema'.format(
                    name
                )
            )

    def get_by_id(self, id):
        """ Return message class for given id """
        try:
            return self.msg_classes_by_id[id]
        except KeyError:
            raise KeyError(
                'No message identified by {} found in the schema'.format(
                    id
                )
            )

    def __init__(self, schema):
        """
        Constructor for message factory, takes schema (dict) as the first
        argument and generates message classes based on the schema.
        """
        keys = list(schema.keys())
        keys.sort()
        next_id = 1

        try:
            schema_length = len(schema)
            self.bytes_needed_for_id = get_bytes_to_represent(schema_length)
            self.id_binary_format = get_symbol_to_represent(schema_length)
        except OverflowError:
            raise ImproperlyConfigured(
                '''Schema can contain the maximum
                number of {} message types.'''.format(sys.maxsize))

        for msg_class_name in keys:
            schema_types_used = schema[msg_class_name]['format'].values()
            mixins = []

            if('enum' in schema_types_used):
                mixins.append(MessageEnumMixin)
            if('string' in schema_types_used):
                mixins.append(MessageStringMixing)

            mixins.append(MessageBase)

            newclass = type(msg_class_name, tuple(mixins), {})

            if('enum' in schema_types_used):
                newclass._enums = {}
                for key in schema[msg_class_name]['enums'].keys():
                    newclass._enums[key] = bidict(
                        schema[msg_class_name]['enums'][key]
                    )

            newclass._binary_format = self.get_binary_format(
                schema[msg_class_name]
            )
            newclass._format = schema[msg_class_name]['format']
            newclass._keys = list(newclass._format)
            newclass._keys.sort()
            # newclass._schema = schema
            newclass._id = next_id
            newclass._length = len(newclass._keys)

            if('string' in schema_types_used):
                newclass._base_binary_format = newclass._binary_format
                newclass._base_binary_length = struct.calcsize(newclass._binary_format.replace('{}s', ''))
                newclass._string_keys = [key for key in newclass.keys
                    if schema[msg_class_name]['format'][key] == 'string'
                ]
                newclass._string_offsets = [
                    struct.calcsize(o) for o in newclass._base_binary_format.split('I{}s')
                ][:-1]
                newclass._string_indexes = [idx for idx, key in enumerate(newclass.keys)
                    if schema[msg_class_name]['format'][key] == 'string'
                ]
            else:
                newclass._struct = struct.Struct(newclass._binary_format)
                newclass._binary_length = newclass._struct.size


            if('enum' in schema_types_used):
                newclass._enum_keys = [key for key in newclass.keys
                    if schema[msg_class_name]['format'][key] == 'enum'
                ]
                newclass._enum_indexes = [idx for idx, key in enumerate(newclass.keys)
                    if schema[msg_class_name]['format'][key] == 'enum'
                ]

            self.msg_classes_by_name[msg_class_name] = newclass
            self.msg_classes_by_id[next_id] = newclass
            next_id += 1



"""
This module contains classes and functions to pack and unpack arbitrary
messages into binary strings based on schema provided.

This module contains the following classes:
    * MessageBase - base class for generated message classes
    * MessageFactory - generates message classes based on schema
    * ImproperlyConfigured - exception for misconfiguration cases

This module contains the following functions:
    * unpack_message - unpack single message from binary
    * unpack_messages - unpack multiple messages from binary
    * pack_message - pack one message into binary
    * pack_messages - packe multiple messages into binary
"""
import sys
import struct
import math
from future.utils import with_metaclass
from builtins import bytes
from bidict import bidict


class MessageBaseMeta(type):
    @property
    def id(cls):
        """
        Unique id generated for this message class based on the schema
        """
        return cls._id

    @property
    def format(cls):
        """
        Format of this message as per schema
        """
        return cls._format

    @property
    def keys(cls):
        """
        Format of this message as per schema
        """
        return cls._keys

    @property
    def enums(cls):
        """
        2-way dictionary for looking up enum values both ways
        """
        return cls._enums

    @property
    def binary_format(cls):
        """
        Binary format of this message class. This cannot be passed to
        struct directly as it needs ot be processed by the class instance
        first to handle dynamic length of a string
        """
        return cls._binary_format

    @property
    def binary_length(cls):
        """
        """
        return cls._binary_length

    @property
    def struct(cls):
        """
        """
        return cls._struct


class MessageBase(with_metaclass(MessageBaseMeta, dict)):
    """
    Base class for a message that can be packed/unpacked
    """
    def __init__(self, *args, **kwargs):
        self.struct = self.__class__._struct
        self.binary_length = self.__class__._binary_length
        super(MessageBase, self).__init__(*args, **kwargs)

    def inflate(self, data):
        self.update(zip(self.__class__.keys, data))

    def deflate(self):
        return [self[key] for key in self.__class__.keys]




class MessageEnumMixin(object):
    @classmethod
    def enum_lookup(cls, enum_name, identifier):
        try:
            enum_field = cls.enums[enum_name]
        except KeyError:
            raise KeyError(
                "Attempted to lookup non-existing enum field {}".format(
                    enum_name
                )
            )
        try:
            return enum_field[identifier]
        except KeyError:
            raise KeyError(
                "No value found for identifier {} in enum {}".format(
                    identifier, enum_name
                )
            )

    @classmethod
    def enum_reverse_lookup(cls, enum_name, value):
        try:
            enum_field = cls.enums[enum_name]
        except KeyError:
            raise KeyError(
                "Attempted to lookup non-existing enum field {}".format(
                    enum_name
                )
            )
        try:
            return enum_field.inv[value]
        except KeyError:
            raise KeyError(
                "No identifier found for value {} in enum {}".format(
                    value, enum_name
                )
            )

    def inflate(self, data):
        super(MessageEnumMixin, self).inflate(data)
        for key in self.__class__._enum_keys:
            self[key] = self.__class__.enums[key].inv[self[key]]

    def deflate(self):
        deflated = super(MessageEnumMixin, self).deflate()
        for i, idx in enumerate(self.__class__._enum_indexes):
            deflated[idx] = self.__class__.enums[self.__class__._enum_keys[i]][deflated[idx]]

        return deflated


class MessageStringMixing(object):
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

        if(args or kwargs):
            self._calc_binary_length()
        else:
            self.binary_length = self.__class__._base_binary_length

    def _calc_binary_length(self):
        self._strings_lengths = [len(bytes(self[key], 'utf-8')) for key in self.__class__._string_keys]
        self.binary_length = self.__class__._base_binary_length + sum(self._strings_lengths)
        self.binary_format = self.__class__._base_binary_format.format(*self._strings_lengths)
        self.struct = struct.Struct(self.binary_format)

    def __setitem__(self, key, value):
        super(MessageStringMixing, self).__setitem__(key, value)
        if(key in self.__class__._string_keys):
            self._calc_binary_length()

    def inflate(self, data):
        super(MessageStringMixing, self).inflate(data)
        for key in self.__class__._string_keys:
            self[key] = self[key].decode('utf-8')

    def deflate(self):
        deflated = super(MessageStringMixing, self).deflate()
        for i, idx in enumerate(self.__class__._string_indexes):
            deflated[i + idx] = bytes(deflated[i + idx], 'utf-8')
            deflated.insert(i + idx, self._strings_lengths[i])

        return deflated



import struct

def unpack_message(packed, factory):
    (msg_id, ) = struct.unpack_from(
        factory.id_binary_format,
        packed, 0
    )
    MsgCls = factory.get_by_id(msg_id)
    try:
        data = MsgCls.struct.unpack(packed)
    except AttributeError as e:
        strings_lengths = [struct.unpack_from('!I', packed, o)[0] for o in MsgCls._string_offsets]
        binary_format = MsgCls._base_binary_format.format(*strings_lengths)
        data = [i for idx, i in enumerate(struct.unpack(binary_format, packed)) if idx not in MsgCls._string_indexes]
    item = MsgCls()
    item.inflate(data[1:])
    return item


def unpack_messages(packed, factory):
    messages = []
    offset = 0
    lng = len(packed)

    while offset < lng:
        (msg_id, ) = struct.unpack_from(
            factory.id_binary_format,
            packed, offset
        )

        MsgCls = factory.get_by_id(msg_id)

        try:
            data = MsgCls.struct.unpack_from(packed, offset)
            offset += MsgCls.binary_length
        except AttributeError as e:
            strings_lengths = [struct.unpack_from('!I', packed, offset + o)[0] for o in MsgCls._string_offsets]
            binary_format = MsgCls._base_binary_format.format(*strings_lengths)
            msg_struct = struct.Struct(binary_format)
            data = [i for idx, i in enumerate(msg_struct.unpack_from(packed, offset)) if idx not in MsgCls._string_indexes]
            offset += msg_struct.size

        item = MsgCls()
        item.inflate(data[1:])
        messages.append(item)

    return messages

def unpack_messages_of_single_type(packed, factory):
    # @TODO: optimise me :)
    return unpack_messages(packed, factory)
# def pack_messages_of_single_type():
#     messages = []
#     offset = 0
#     lng = len(packed)
#     (msg_id, ) = struct.unpack_from(
#         factory.id_binary_format,
#         packed, 0
#     )
#     offset += factory.bytes_needed_for_id
#     MsgCls = factory.get_by_id(msg_id)

#     while offset < lng:
#         try:
#             data = MsgCls.struct.unpack_from(packed, offset)
#         except Exception as e:
#             strings_lengths = [struct.unpack_from('!I', packed, offset + o)[0] for o in MsgCls._string_offsets]
#             binary_format = MsgCls._base_binary_length.format(strings_lengths)
#             msg_struct = struct.Struct(binary_format)
#             data = msg_struct.unpack_from(packed, offset)
#             offset += msg_struct.size - factory.bytes_needed_for_id

#         item = MsgCls(*data[1:])
#         messages.append(item)

#     return messages




import sys
import math

byte_number_to_symbol = {
    1: 'B',
    2: 'H',
    3: 'I',
    4: 'I',
    5: 'Q',
    6: 'Q',
    7: 'Q',
    8: 'Q'
}

def get_bytes_to_represent(number):
    return math.ceil(math.log(number + 1, 2) / 8)

def get_symbol_to_represent(number):
    if(number > sys.maxsize or number > 1.8446744073709552e+19):
        raise OverflowError(
            "Unable to represent number {} in packed structure".format(
                number
            )
        )
    return get_binary_format_symbol(get_bytes_to_represent(number))

def get_binary_format_symbol(bytes_needed):
    return byte_number_to_symbol[bytes_needed]



import io
from ctypes import create_string_buffer

def pack_message(message):
    return message.struct.pack(message.__class__.id, *message.deflate())

def pack_messages(messages):
    lng = 0
    for msg in messages:
        lng += msg.binary_length

    buf = create_string_buffer(lng)
    offset = 0

    for msg in messages:
        msg.struct.pack_into(buf, offset, msg.__class__.id, *msg.deflate())
        offset += msg.binary_length

    return buf.raw


def pack_messages_of_single_type(messages):
    buf = create_string_buffer(messages[0].binary_length * len(messages))
    offset = 0

    for msg in messages:
        msg.struct.pack_into(buf, offset, msg.__class__.id, *msg.deflate())
        offset += msg.binary_length

    return buf.raw



from .factory import MessageFactory
from .packers import pack_message, pack_messages, pack_messages_of_single_type
from .unpackers import unpack_message, unpack_messages, unpack_messages_of_single_type
